#include <cuda_runtime.h>

#define BLOCK_SIZE_X	32
#define BLOCK_SIZE_Y	16
#define BLOCKS_X		32
#define BLOCKS_Y		64

#define VOXELS_X		1024
#define VOXELS_Y		1024
#define VOXELS_Z		128

#define RCP_VOXELS_X	0.0009765625f;	// 1/1024
#define RCP_VOXELS_Y	0.0009765625f;	// 1/1024
#define RCP_VOXELS_Z	0.0078125f;		// 1/128

texture<float, 2, cudaReadModeElementType> tex1;
texture<float, 2, cudaReadModeElementType> tex2;
texture<float, 2, cudaReadModeElementType> tex3;
texture<float, 2, cudaReadModeElementType> tex4;

__global__ void kernelZero(float *sptr, float *eptr)
{
	unsigned int x = blockIdx.x * BLOCK_SIZE_X + threadIdx.x;
	unsigned int y = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y;

	unsigned int i = (blockIdx.y * BLOCKS_X + blockIdx.x) * BLOCK_SIZE_X * BLOCK_SIZE_Y + (threadIdx.y * BLOCK_SIZE_X) + threadIdx.x;
	
	volatile float *ptr = (float *)((unsigned int)sptr + i * sizeof(float));
	
	while(ptr < eptr)
	{
		*ptr = 0.0f;
		
		ptr += VOXELS_X * VOXELS_Y;
	}
}

extern "C" void cudaZero(float *slices)
{
    dim3 dimBlock(BLOCK_SIZE_X, BLOCK_SIZE_Y, 1);
    dim3 dimGrid(BLOCKS_X, BLOCKS_Y);

	kernelZero<<< dimGrid, dimBlock, 0 >>>(slices, slices + VOXELS_X * VOXELS_Y * VOXELS_Z);
}

__global__ void kernelSample4(float *sptr, float *eptr)
{
	unsigned int x = blockIdx.x * BLOCK_SIZE_X + threadIdx.x;
	unsigned int y = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y;

	unsigned int i = (blockIdx.y * BLOCKS_X + blockIdx.x) * BLOCK_SIZE_X * BLOCK_SIZE_Y + (threadIdx.y * BLOCK_SIZE_X) + threadIdx.x;
	
	float u = (x + 0.5f) * RCP_VOXELS_X;
	float v = 0.5f * RCP_VOXELS_Z;
	
	volatile float *ptr = (float *)((unsigned int)sptr + i * sizeof(float));
	
	#pragma unroll 1
	while(ptr < eptr)
	{
		float accum = *ptr;
		
		accum += tex2D(tex1, u, v);
		accum += tex2D(tex2, u, v);
		accum += tex2D(tex3, u, v);
		accum += tex2D(tex4, u, v);

		v += RCP_VOXELS_Z;
	
		*ptr = accum;

		ptr += VOXELS_X * VOXELS_Y;
	}
}

extern "C" void cudaSample4(float *slices, cudaArray *ptr1, cudaArray *ptr2, cudaArray *ptr3, cudaArray *ptr4)
{
	cudaChannelFormatDesc channelDesc1 = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	tex1.addressMode[0] = cudaAddressModeClamp;
	tex1.addressMode[1] = cudaAddressModeClamp;
	tex1.filterMode = cudaFilterModeLinear;
	tex1.normalized = true;

	cudaBindTextureToArray(tex1, ptr1, channelDesc1);

	cudaChannelFormatDesc channelDesc2 = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	tex2.addressMode[0] = cudaAddressModeClamp;
	tex2.addressMode[1] = cudaAddressModeClamp;
	tex2.filterMode = cudaFilterModeLinear;
	tex2.normalized = true;

	cudaBindTextureToArray(tex2, ptr2, channelDesc2);

	cudaChannelFormatDesc channelDesc3 = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	tex3.addressMode[0] = cudaAddressModeClamp;
	tex3.addressMode[1] = cudaAddressModeClamp;
	tex3.filterMode = cudaFilterModeLinear;
	tex3.normalized = true;

	cudaBindTextureToArray(tex3, ptr3, channelDesc3);

	cudaChannelFormatDesc channelDesc4 = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);

	tex4.addressMode[0] = cudaAddressModeClamp;
	tex4.addressMode[1] = cudaAddressModeClamp;
	tex4.filterMode = cudaFilterModeLinear;
	tex4.normalized = true;

	cudaBindTextureToArray(tex4, ptr4, channelDesc4);

	dim3 dimBlock(BLOCK_SIZE_X, BLOCK_SIZE_Y, 1);
	dim3 dimGrid(BLOCKS_X, BLOCKS_Y);

	kernelSample4<<< dimGrid, dimBlock, 0 >>>(slices, slices + VOXELS_X * VOXELS_Y * VOXELS_Z);

	cudaUnbindTexture(tex1);
	
	cudaUnbindTexture(tex2);
	
	cudaUnbindTexture(tex3);
	
	cudaUnbindTexture(tex4);
}
